@@ -33,7 +33,9 @@ const std::array<std::string, Protocol::s_numberOfProtocols> Protocol::s_names =
   "DNSCryptUDP",
   "DNSCryptTCP",
   "DoT",
-  "DoH"};
+  "DoH",
+  "DoQ",
+  "DoH3"};
 
 const std::array<std::string, Protocol::s_numberOfProtocols> Protocol::s_prettyNames = {
   "Do53 UDP",
@@ -41,7 +43,9 @@ const std::array<std::string, Protocol::s_numberOfProtocols> Protocol::s_prettyN
   "DNSCrypt UDP",
   "DNSCrypt TCP",
   "DNS over TLS",
-  "DNS over HTTPS"};
+  "DNS over HTTPS",
+  "DNS over QUIC",
+  "DNS over HTTP/3"};
 
 Protocol::Protocol(const std::string& s)
 {
@@ -79,6 +83,11 @@ bool Protocol::isUDP() const
   return d_protocol == DoUDP || d_protocol == DNSCryptUDP;
 }
 
+bool Protocol::isEncrypted() const
+{
+  return d_protocol != DoUDP && d_protocol != DoTCP;
+}
+
 uint8_t Protocol::toNumber() const
 {
   return static_cast<uint8_t>(d_protocol);
